import json
import pathlib


class Config:
    def __init__(self, root_dir=".", projects=None, **kwargs):
        if projects is None:
            projects = []
        self.root_dir = root_dir
        self.projects = projects

    @classmethod
    def from_file(cls, realm_json_file):
        p = pathlib.Path(realm_json_file)
        with open(realm_json_file) as f:
            cfg_json = json.load(f)

            return cls(root_dir=str(p.parent), **cfg_json)



import fnmatch
import platform
from typing import List

from realm.entities import RealmContext

IS_WINDOWS = any(platform.win32_ver())


class ScopeFilter:
    @classmethod
    def apply(cls, ctx: RealmContext, scopes: List[str]):
        if not scopes:
            return

        filtered = []
        for scope in scopes:
            if IS_WINDOWS:
                scope = scope.replace("^", "*")
            f = {p for p in ctx.projects if fnmatch.fnmatch(p.name, scope)}
            filtered.append(f)

        ctx.projects = cls._union(filtered)

    @classmethod
    def _union(cls, filtered):
        result = filtered[0]
        for f in filtered:
            result = result.union(f)
        return list(result)



import shlex
import subprocess


class ChildProcess:
    FORCE_CAPTURE = False
    CAPTURE_PARAMS = {"stdout": subprocess.PIPE}

    @classmethod
    def run(cls, command, **kwargs) -> str:
        params = dict(cls.CAPTURE_PARAMS)
        params.update(kwargs)

        if cls.FORCE_CAPTURE:
            params.update(cls.CAPTURE_PARAMS)

        if not kwargs.get("shell"):
            command = shlex.split(command, posix=False)
        p = subprocess.Popen(command, **params)
        out, err = p.communicate()
        if isinstance(out, bytes):
            out = out.decode("utf-8", "ignore")

        if isinstance(err, bytes):
            err = err.decode("utf-8", "ignore")

        if p.returncode != 0:
            raise RuntimeError(
                f"Failed running command: {command};\nstdout: {out};\nstderr: {err};"
            )

        return out



from typing import List

from .config import Config
from .graph import Graph
from .project import Project


class RealmContext:
    def __init__(self, config: Config, projects: List[Project]):
        self.config = config
        self.projects = projects
        self.all_projects = projects
        self.dependency_graph = Graph(projects)



import json
import pathlib

from realm.cli.realm_command import RealmCommand
from realm.utils.child_process import ChildProcess

INIT_DATA = {"projects": ["packages/*"]}


class InitCommand(RealmCommand[dict]):
    NAME = "init"
    HELP_MESSAGE = """
    Initializes a new realm repo
    """

    def run(self):
        self.create_realm_json()
        self.create_stub_package()

    @staticmethod
    def create_realm_json():
        with open("realm.json", "w") as f:
            json.dump(INIT_DATA, f, indent=2)

    @staticmethod
    def create_stub_package():
        packages_dir = pathlib.Path("packages")
        if not packages_dir.is_dir():
            packages_dir.mkdir(parents=True)
        ChildProcess.run("poetry new pgk --src", cwd=packages_dir.absolute())



import glob
import json
import logging
import os
import pathlib
import sys
from typing import List

import click

from realm.entities import Config, Project, RealmContext
from realm.log import logger
from realm.version import __version__

from .commands.build import BuildCommand
from .commands.init import InitCommand
from .commands.install import InstallCommand
from .commands.ls import LsCommand
from .commands.run import RunCommand
from .commands.task import TaskCommand
from .core.params import GlobalOption
from .realm_click_types import RealmClickGroup

CONFIG_FILE = "realm.json"


class Application:
    @classmethod
    def create(cls):
        grp = RealmClickGroup(callback=cls.init_context, params=cls.global_options())

        grp.add_command(InitCommand)
        grp.add_command(LsCommand)
        grp.add_command(RunCommand)
        grp.add_command(InstallCommand)
        grp.add_command(TaskCommand)
        grp.add_command(BuildCommand)

        return grp

    @classmethod
    def global_options(cls):
        return [
            GlobalOption(
                ["--version", "-V"],
                is_flag=True,
                callback=cls.print_version,
                help="Display realm version",
            ),
            GlobalOption(
                ["--verbose", "-v"],
                show_default=True,
                count=True,
                help="Sets the verbosity level",
            ),
            GlobalOption(
                ["--parallelism", "-p"],
                type=click.INT,
                show_default=True,
                default=1,
                help="Sets the parallelism for the command (if supported)",
            ),
            GlobalOption(
                ["--since"],
                help="Includes only projects affected by a change since the specified reference",
            ),
            GlobalOption(
                ["--all"],
                is_flag=True,
                help="Includes all projects if no projects were changed when using the --since filter",
            ),
            GlobalOption(
                ["--scope"],
                type=click.STRING,
                help="Includes only projects that match the given wildcard pattern",
                multiple=True,
            ),
            GlobalOption(
                ["--ignore"],
                type=click.STRING,
                help="Filters out projects that match the given wildcard pattern",
                multiple=True,
            ),
            GlobalOption(
                ["--match"],
                type=click.STRING,
                help="Filters by a field specified in `pyproject.toml`",
                multiple=True,
            ),
        ]

    @staticmethod
    def print_version(ctx, _, value):
        if value:
            msg = "Realm {}".format(click.style(__version__, fg="yellow"))
            click.echo(msg)

            sys.exit(0)

    @staticmethod
    @click.pass_context
    def init_context(ctx, **kwargs):
        Application._set_verbosity(kwargs.get("verbose", 0))
        cfg = Application.read_config()
        projects = Application.get_projects(cfg)
        ctx.obj = RealmContext(config=cfg, projects=projects)

    @classmethod
    def read_config(cls):
        current = pathlib.Path(os.getcwd())
        current = current.joinpath(CONFIG_FILE)
        possible_paths = [x.joinpath(CONFIG_FILE) for x in current.parents]
        for p in possible_paths:
            if p.exists():
                with p.open() as f:
                    cfg_json = json.load(f)
                    return Config(root_dir=str(p.parent), **cfg_json)

        return Config()

    @classmethod
    def get_projects(cls, cfg: Config) -> List[Project]:
        tmp_candidates = [
            glob.glob(os.path.join(cfg.root_dir, p)) for p in cfg.projects
        ]
        candidates = []
        for lst in tmp_candidates:
            for c in lst:
                candidates.append(os.path.abspath(c))

        projects = []
        for path in candidates:
            if os.path.isdir(path) and "pyproject.toml" in os.listdir(path):
                projects.append(Project(source_dir=path, root_dir=cfg.root_dir))

        return projects

    @staticmethod
    def _set_verbosity(verbose: int):
        if verbose == 0:
            logger.setLevel(logging.CRITICAL)
        elif verbose == 1:
            logger.setLevel(logging.WARNING)
        elif verbose == 2:
            logger.setLevel(logging.INFO)
        elif verbose >= 3:
            logger.setLevel(logging.DEBUG)


cli = Application.create()



import fnmatch
import platform
from typing import List

from realm.entities import RealmContext

IS_WINDOWS = any(platform.win32_ver())


class IgnoreFilter:
    @classmethod
    def apply(cls, ctx: RealmContext, ignores: List[str]):
        if not ignores:
            return

        filtered = []
        for ignore in ignores:
            if IS_WINDOWS:
                ignore = ignore.replace("^", "*")
            f = {p for p in ctx.projects if not fnmatch.fnmatch(p.name, ignore)}
            filtered.append(f)

        ctx.projects = cls._intersect(filtered)

    @classmethod
    def _intersect(cls, filtered):
        result = filtered[0]
        for f in filtered:
            result = result.intersection(f)
        return list(result)



import os
from functools import cached_property
from pathlib import Path

import toml

from realm.log import logger
from realm.utils.child_process import ChildProcess

PYPROJECT_FILE = "pyproject.toml"


class Project:
    def __init__(self, source_dir, root_dir):
        self.source_dir = Path(source_dir)
        self.name = os.path.basename(source_dir)
        self.relative_path = source_dir[len(root_dir) :].lstrip(os.sep)

        self.pyproject_toml_path = self.source_dir.joinpath(PYPROJECT_FILE)
        self.pyproject = toml.load(self.pyproject_toml_path)

    @property
    def version(self) -> str:
        return self.pyproject["tool"]["poetry"]["version"]

    @property
    def package_name(self) -> str:
        return self.pyproject["tool"]["poetry"]["name"]

    def extract_field(self, toml_path: str):
        parts = toml_path.split(".")
        current = self.pyproject
        for p in parts:
            current = current.get(p, {})
        if bool(current):
            return current
        return None

    @cached_property
    def dependencies(self):
        tool_poetry = self.pyproject["tool"]["poetry"]
        all_dependencies = {}
        for _, group_value in tool_poetry.get("group", {}).items():
            all_dependencies.update(group_value.get("dependencies", {}))

        all_dependencies.update(tool_poetry.get("dev-dependencies", {}))
        all_dependencies.update(tool_poetry["dependencies"])

        return all_dependencies

    def has_task(self, task_name) -> bool:
        tasks = self.pyproject["tool"].get("poe", {}).get("tasks", {})
        return task_name in tasks

    def execute_cmd(self, cmd, **kwargs):
        full_cmd = cmd
        env = self._create_env()

        try:
            params = {
                "stdout": None,
                "stderr": None,
                "shell": True,
                "env": env,
                "cwd": self.source_dir,
            }
            params.update(kwargs)
            return ChildProcess.run(full_cmd, **params)
        except RuntimeError as e:
            msg = f"{e!s}\nproject: {os.path.basename(self.source_dir)}"
            raise RuntimeError(msg) from e

    def is_dependent_on(self, other: "Project") -> bool:
        dependency = self.dependencies.get(other.package_name)
        return self.is_dependency(other, dependency)

    def is_dependency(self, other: "Project", dependency: dict) -> bool:
        if dependency is None:
            return False
        if not isinstance(dependency, dict):
            # Path dependency is a dict
            return False
        dependency_path = dependency.get("path")
        if dependency_path is None:
            return False
        dependency_path = dependency_path.replace("/", os.sep)
        resolved_dep_path = (
            self.source_dir.joinpath(dependency_path).resolve().absolute()
        )
        logger.debug(
            f"Checking if {self} is dependent on {other} with path {dependency_path}"
        )

        eq = resolved_dep_path == other.source_dir.resolve().absolute()
        logger.debug(
            f"Resolved path: {resolved_dep_path} == {other.source_dir.resolve().absolute()}? {eq}"
        )
        return eq

    def __repr__(self):
        return self.package_name

    def _create_env(self):
        env = dict(os.environ)
        current_venv = os.getenv("VIRTUAL_ENV", os.getenv("CONDA_PREFIX"))
        if current_venv:
            path_env = os.environ.get("PATH", "")
            # Remove venv from path
            env["PATH"] = ":".join(
                [e for e in path_env.split(":") if current_venv not in e]
            )
            env.pop("VIRTUAL_ENV", None)
            env.pop("CONDA_PREFIX", None)
        env["REALM_PROJECT_NAME"] = self.name
        env["REALM_PROJECT_PATH"] = str(self.source_dir)

        return env



import sys

import click

from realm.cli.realm_command import RealmCommand


class TaskCommand(RealmCommand[dict]):
    NAME = "task"
    HELP_MESSAGE = """
    Runs a poe task on all projects containing this task
    """
    PARAMS = [click.Argument(["task_name"], type=click.STRING)]

    def _filter_projects(self):
        super()._filter_projects()
        task_name = self.params["task_name"]
        self.ctx.projects = [p for p in self.ctx.projects if p.has_task(task_name)]

    def run(self):
        task_name = self.params["task_name"]
        failed_projects = []

        self._run_in_pool(
            self.execute_task, task_name=task_name, failed_projects=failed_projects
        )

        if any(failed_projects):
            names = [p.name for p in failed_projects]
            formatted_names = ", ".join(names)
            msg = f"Task {task_name} failed for projects: {formatted_names}"
            click.echo(msg, err=True)
            sys.exit(1)

    @staticmethod
    def execute_task(project, task_name, failed_projects):
        click.echo(f"Running task {task_name} for: {project.name}")
        try:
            out = project.execute_cmd(f"poetry run poe {task_name}")
            if out:
                click.echo(out)
        except RuntimeError as e:
            click.echo(e, err=True)
            failed_projects.append(project)



def __get_runtime_version():
    import importlib.metadata

    try:
        return importlib.metadata.version("realm")
    except importlib.metadata.PackageNotFoundError:
        return None


__version__ = __get_runtime_version()



import click

from realm.cli.realm_command import RealmCommand


class LsCommand(RealmCommand[dict]):
    NAME = "ls"
    HELP_MESSAGE = """
    Prints all projects managed
    """
    PARAMS = [
        click.Option(
            ["--paths"], help="Prints relative paths of the projects", is_flag=True
        )
    ]

    def run(self):
        for p in self.ctx.projects:
            if self.params.get("paths"):
                click.echo(f"{p.relative_path}")
            else:
                click.echo(f"{p.name}@{p.version}")



import abc
import re
import sys
from typing import Generic, List, Optional, Type, TypeVar

import click

T = TypeVar("T")


def _strip_indent(s):
    if s is None:
        return None
    pattern = re.compile(r"^[ \t]*(?=\S)", re.MULTILINE)
    indent = min(len(spaces) for spaces in pattern.findall(s))

    if not indent:
        return s

    return re.sub(re.compile(r"^[ \t]{%s}" % indent, re.MULTILINE), "", s)


class BaseCommand(Generic[T]):
    CLICK_COMMAND_CLS = click.Command
    NAME: str = ""
    HELP_MESSAGE: Optional[str] = None
    PARAMS: List[click.Parameter] = []
    DEPRECATED: bool = False
    SCHEMA: Type[T] = dict

    def __init__(self, ctx, **kwargs):
        self.ctx = ctx
        self.params: T = self.SCHEMA(**kwargs)
        self._params: dict = kwargs

    @classmethod
    def to_command(cls):
        def callback_fn(ctx, **kwargs):
            cmd = cls(ctx=ctx.obj, **kwargs)
            code = cmd.run()
            if code is not None:
                sys.exit(code)

        return cls.CLICK_COMMAND_CLS(
            name=cls.NAME,
            params=cls.PARAMS,
            callback=click.pass_context(callback_fn),
            help=_strip_indent(cls.HELP_MESSAGE),
            deprecated=cls.DEPRECATED,
        )

    @abc.abstractmethod
    def run(self):  # type: Optional[int]
        """
        Runs the command
        :return: Exit code
        """
        raise NotImplementedError()



import os
from typing import List

from realm.entities import RealmContext
from realm.utils.child_process import ChildProcess


class SinceFilter:
    @classmethod
    def apply(cls, ctx: RealmContext, since):
        if not since:
            return

        changed_projects = cls.get_changed_projects(ctx, since)
        affected = set()
        nodes = list(changed_projects)

        while any(nodes):
            proj = nodes.pop()
            if proj not in affected:
                affected.add(proj)
                proj_affected = ctx.dependency_graph.projects_affected[proj]
                nodes.extend(proj_affected)

        ctx.projects = affected

    @classmethod
    def get_changed_projects(cls, ctx: RealmContext, since):
        changed_files = ChildProcess.run(f"git diff --name-only {since}")
        git_root = ChildProcess.run("git rev-parse --show-toplevel")
        git_root = git_root.strip() if git_root else ""

        relative_realm_repo_path = ctx.config.root_dir[len(git_root) :].lstrip(os.sep)

        changed_files = cls.parse_changed_files(changed_files)

        ctx.projects.sort(key=lambda p: len(str(p.source_dir)), reverse=True)
        changed = set()

        for file in changed_files:
            for proj in ctx.projects:
                project_path = os.path.join(
                    relative_realm_repo_path, proj.relative_path
                ).rstrip(os.path.sep)
                project_path += os.path.sep
                if file.startswith(project_path):
                    changed.add(proj)
                    continue
        return changed

    @classmethod
    def parse_changed_files(cls, changed_files: str) -> List[str]:
        def handle_windows(p: str):
            return p.replace("/", os.sep)

        changed_files = changed_files if changed_files else ""
        return [handle_windows(f.strip()) for f in changed_files.split("\n") if f]



import shlex
import sys

import click

from realm.cli.realm_command import RealmCommand
from realm.entities import Project


class RunCommand(RealmCommand[dict]):
    NAME = "run"
    HELP_MESSAGE = """
    Executes a command on all projects
    """
    PARAMS = [click.Argument(["command"], type=click.STRING, nargs=-1)]

    def run(self):
        full_cmd = self.get_full_cmd()
        self.logger.debug(f"Running command: {full_cmd}")
        try:
            self._run_in_pool(self._run_cmd, cmd=full_cmd)
        except Exception as e:
            click.echo(e, err=True)
            sys.exit(1)

    def get_full_cmd(self):
        cmd = self.params.get("command")
        full_cmd = shlex.join(cmd)
        return full_cmd

    @staticmethod
    def _run_cmd(project: Project, cmd: str):
        try:
            out = project.execute_cmd(cmd)
            if out:
                # used only for tests :(
                click.echo(out)
        except Exception as e:
            click.echo(e, err=True)
            sys.exit(1)



import fnmatch
from typing import List

from realm.entities import Project, RealmContext


class MatchFilter:
    @classmethod
    def apply(cls, ctx: RealmContext, matches: List[str]):
        if not matches:
            return

        filtered = []
        for match in matches:
            field, _, value = match.partition("=")
            f1 = f"tool.realm.{field}"
            f2 = f"tool.{field}"
            f = {
                p
                for p in ctx.projects
                if cls._matches(p, f1, value) or cls._matches(p, f2, value)
            }
            filtered.append(f)

        ctx.projects = cls._union(filtered)

    @classmethod
    def _matches(cls, project: Project, field: str, pattern: str) -> bool:
        value = project.extract_field(field)
        if not value:
            return False
        return fnmatch.fnmatch(value, pattern)

    @classmethod
    def _union(cls, filtered):
        result = filtered[0]
        for f in filtered:
            result = result.union(f)
        return list(result)



import os
from abc import ABC
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Callable

from realm.entities import Project, RealmContext
from realm.filters.ignore import IgnoreFilter
from realm.filters.match import MatchFilter
from realm.filters.scope import ScopeFilter
from realm.filters.since import SinceFilter
from realm.log import logger
from realm.utils import await_all

from .core.base_command import BaseCommand, T
from .realm_click_types import RealmClickCommand


class RealmCommand(BaseCommand[T], ABC):
    CLICK_COMMAND_CLS = RealmClickCommand

    def __init__(self, ctx: RealmContext, **kwargs):
        super().__init__(ctx, **kwargs)
        self.logger = logger
        # Just for the type annotation
        self.ctx = ctx
        self.pool = ThreadPoolExecutor(self._params.get("parallelism", 1))
        self._filter_projects()
        os.environ["REALM_ROOT_PATH"] = self.ctx.config.root_dir

    def _filter_projects(self):
        SinceFilter.apply(ctx=self.ctx, since=self._params.get("since"))

        include_all_when_empty = self._params.get("all")
        if include_all_when_empty and not any(self.ctx.projects):
            self.ctx.projects = self.ctx.all_projects

        ScopeFilter.apply(ctx=self.ctx, scopes=self._params.get("scope"))

        IgnoreFilter.apply(ctx=self.ctx, ignores=self._params.get("ignore"))

        MatchFilter.apply(ctx=self.ctx, matches=self._params.get("match"))

    def _run_in_pool(self, func: Callable[[Project, Any], Any], *args, **kwargs):
        projects = set(self.ctx.projects)
        for level in self.ctx.dependency_graph.topology:
            futures = [
                self.pool.submit(func, proj, *args, **kwargs)
                for proj in level
                if proj in projects
            ]
            await_all(futures)



import sys

import click

from realm.cli.realm_command import RealmCommand
from realm.entities.project import Project


class InstallCommand(RealmCommand[dict]):
    NAME = "install"
    HELP_MESSAGE = """
    Executes "poetry install" on all projects
    """

    def run(self):
        self._run_in_pool(self._install)

    @staticmethod
    def _install(project: Project):
        try:
            out = project.execute_cmd("poetry install")
            if out:
                # used only for tests :(
                click.echo(out)
        except Exception as e:
            click.echo(e, err=True)
            sys.exit(1)



import shutil
import sys
from pathlib import Path

import click
import toml

from realm.cli.realm_command import RealmCommand
from realm.entities.project import Project


class BuildCommand(RealmCommand[dict]):
    NAME = "build"
    HELP_MESSAGE = """
    Runs poetry build on all projects, it will replace path-dependencies with their current version in the workspace
    """

    def run(self):
        self._run_in_pool(self._build)

    def _build(self, project: Project):
        bak_file = project.pyproject_toml_path.with_name("_bak_realm_pyproject.toml")

        try:
            self._prep_build(project, bak_file)

            out = project.execute_cmd("poetry build")
            if out:
                click.echo(out)
        except Exception as e:
            click.echo(e, err=True)
            sys.exit(1)
        finally:
            if bak_file.is_file():
                shutil.copy2(bak_file, project.pyproject_toml_path)
                bak_file.unlink()

    def _prep_build(self, project: Project, bak_file: Path):
        tmp_pyproject = dict(project.pyproject)

        dependencies = self.ctx.dependency_graph.project_deps[project]
        if not any(dependencies):
            return

        shutil.copy2(project.pyproject_toml_path, bak_file)
        for dep_group, dep_name, dep_version in self._iterate_dependencies(
            tmp_pyproject
        ):
            for proj_dep in dependencies:
                if project.is_dependency(proj_dep, dep_version):
                    dep_group[dep_name] = f"^{proj_dep.version}"
                    break

        with project.pyproject_toml_path.open("w") as f:
            toml.dump(tmp_pyproject, f)

    @staticmethod
    def _iterate_dependencies(pyporject: dict):
        def iterate_dict(d: dict):
            for k, v in d.items():
                yield d, k, v

        tool_poetry = pyporject["tool"]["poetry"]
        for _, group_value in tool_poetry.get("group", {}).items():
            grp = group_value.get("dependencies", {})
            yield from iterate_dict(grp)
        yield from iterate_dict(tool_poetry.get("dev-dependencies", {}))
        yield from iterate_dict(tool_poetry.get("dependencies", {}))

    @staticmethod
    def _is_path_dependency():
        pass



import logging

from colorama import Fore


class ColorFormatter(logging.Formatter):
    grey = Fore.LIGHTBLACK_EX
    yellow = Fore.YELLOW
    red = Fore.RED
    reset = Fore.RESET
    format = "%(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: red + format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def __create_logger():
    import logging

    logger = logging.getLogger("realm")
    handler = logging.StreamHandler()
    handler.setFormatter(ColorFormatter())
    logger.addHandler(handler)
    return logger


logger = __create_logger()



import click

from realm.cli.formatter import RealmFormatHelpMixin

from .core.group import Group


class RealmClickCommand(RealmFormatHelpMixin, click.Command):
    pass


class RealmClickGroup(RealmFormatHelpMixin, Group):
    def format_options(self, ctx, formatter):
        RealmFormatHelpMixin.format_options(self, ctx, formatter)
        self.format_commands(ctx, formatter)



from itertools import combinations
from typing import Dict, List, Set, Tuple

from .project import Project

ProjectDependencies = Dict[Project, Set[Project]]


class Graph:
    def __init__(self, projects: List[Project]):
        self.projects = projects
        self.project_deps, self.projects_affected = self._build_project_deps(projects)
        self.topology = self._build_topology(self.project_deps)

    @classmethod
    def _build_project_deps(
        cls, projects: List[Project]
    ) -> Tuple[ProjectDependencies, ProjectDependencies]:
        project_deps = {p: set() for p in projects}
        projects_affected = {p: set() for p in projects}

        proj_a: Project
        proj_b: Project
        for proj_a, proj_b in combinations(projects, 2):
            if proj_a.is_dependent_on(proj_b):
                project_deps[proj_a].add(proj_b)
                projects_affected[proj_b].add(proj_a)
            elif proj_b.is_dependent_on(proj_a):
                project_deps[proj_b].add(proj_a)
                projects_affected[proj_a].add(proj_b)

        return project_deps, projects_affected

    @classmethod
    def _build_topology(cls, project_deps: ProjectDependencies) -> List[Set[Project]]:
        project_deps = dict(project_deps)
        res = [set()]
        seen_deps = set()
        while any(project_deps):
            seen_deps = seen_deps.union(res[-1])
            new = set()
            for proj, deps in list(project_deps.items()):
                if not any(deps - seen_deps):
                    new.add(proj)
                    project_deps.pop(proj)
            res.append(new)
        return res[1:]
